"""A domain to register in sphinx.

This is required for any directive/role names using `:`.
"""
from typing import List

from sphinx.domains import Domain

from . import get_glue_directives, get_glue_roles


class NbGlueDomain(Domain):
    """A sphinx domain for defining glue roles and directives."""

    name = "glue"
    label = "NotebookGlue"

    # data version, bump this when the format of self.data changes
    data_version = 1

    directives = get_glue_directives(prefix="")
    roles = get_glue_roles(prefix="")

    def merge_domaindata(self, docnames: List[str], otherdata: dict) -> None:
        pass



"""Utilities for working with docutils and sphinx.

We intentionally do no import sphinx in this module,
in order to allow docutils-only use without sphinx installed.
"""
from ast import literal_eval
import dataclasses as dc
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from docutils import nodes

from myst_nb._compat import findall
from myst_nb.core.loggers import DocutilsDocLogger, SphinxDocLogger
from myst_nb.core.render import MimeData, NbElementRenderer, get_mime_priority

if TYPE_CHECKING:
    from sphinx.environment import BuildEnvironment


def is_sphinx(document) -> bool:
    """Return True if we are in sphinx, otherwise docutils."""
    return hasattr(document.settings, "env")


def glue_warning(
    message: str, document: nodes.document, line: int
) -> nodes.system_message:
    """Create a warning related to glue functionality."""
    logger: Union[DocutilsDocLogger, SphinxDocLogger]
    if is_sphinx(document):
        logger = SphinxDocLogger(document)
    else:
        logger = DocutilsDocLogger(document)
    logger.warning(message, line=line, subtype="glue")
    return nodes.system_message(
        message,
        type="WARNING",
        level=2,
        line=line,
        source=document["source"],
    )


def set_source_info(node: nodes.Node, source: str, line: int) -> None:
    """Set the source info for a node and its descendants."""
    for _node in findall(node)(include_self=True):
        _node.source = source
        _node.line = line


@dc.dataclass()
class RetrievedData:
    """A class to store retrieved mime data."""

    data: Dict[str, Any]
    metadata: Dict[str, Any]
    nb_renderer: NbElementRenderer


class RetrievalError(Exception):
    """An error occurred while retrieving the glue data."""


def retrieve_glue_data(document: nodes.document, key: str) -> RetrievedData:
    """Retrieve the glue data from a specific document."""
    msg = f"No key {key!r} found in glue data for this document."
    if "nb_renderer" not in document:
        raise RetrievalError(msg)
    nb_renderer: NbElementRenderer = document["nb_renderer"]
    resources = nb_renderer.get_resources()
    if "glue" not in resources:
        raise RetrievalError(msg)

    if key not in resources["glue"]:
        raise RetrievalError(msg)

    return RetrievedData(
        data=resources["glue"][key]["data"],
        metadata=resources["glue"][key].get("metadata", {}),
        nb_renderer=nb_renderer,
    )


def render_glue_output(
    data: RetrievedData,
    document: nodes.document,
    line: int,
    source: str,
    *,
    inline: bool = False,
    render: Optional[Dict[str, Any]] = None,
) -> List[nodes.Node]:
    """Retrive the notebook output data for this glue key,
    then return the docutils/sphinx nodes relevant to this data.

    :param key: The glue key to retrieve.
    :param document: The current docutils document.
    :param line: The current source line number of the directive or role.
    :param source: The current source path or description.
    :param inline: Whether to render the output as inline (or block).
    :param render: Cell-level render metadata

    :returns: A tuple of (was the key found, the docutils/sphinx nodes).
    """
    cell_metadata = {}
    if render:
        cell_metadata[data.nb_renderer.config.cell_metadata_key] = render
    if is_sphinx(document):
        _nodes = _render_output_sphinx(
            data.nb_renderer,
            data.data,
            cell_metadata,
            data.metadata,
            source,
            line,
            inline,
        )
    else:
        _nodes = _render_output_docutils(
            data.nb_renderer,
            data.data,
            cell_metadata,
            data.metadata,
            document,
            line,
            inline,
        )
    # TODO rendering should perhaps return if it succeeded explicitly,
    # and whether system_messages or not (required for roles)
    return _nodes


def _render_output_docutils(
    nb_renderer: NbElementRenderer,
    data: Dict[str, Any],
    cell_metadata: Dict[str, Any],
    output_metadata: Dict[str, Any],
    document: nodes.document,
    line: int,
    inline=False,
) -> List[nodes.Node]:
    """Render the output in docutils (select mime priority directly)."""
    mime_priority = get_mime_priority(
        nb_renderer.config.builder_name,
        nb_renderer.config.mime_priority_overrides,
    )
    try:
        mime_type = next(x for x in mime_priority if x in data)
    except StopIteration:
        return [
            glue_warning(
                "No output mime type found from render_priority",
                document,
                line,
            )
        ]
    else:
        mime_data = MimeData(
            mime_type,
            data[mime_type],
            cell_metadata=cell_metadata,
            output_metadata=output_metadata,
            line=line,
        )
        if inline:
            return nb_renderer.render_mime_type_inline(mime_data)
        return nb_renderer.render_mime_type(mime_data)


def _render_output_sphinx(
    nb_renderer: NbElementRenderer,
    data: Dict[str, Any],
    cell_metadata: Dict[str, Any],
    output_metadata: Dict[str, Any],
    source: str,
    line: int,
    inline=False,
) -> List[nodes.Node]:
    """Render the output in sphinx (defer mime priority selection)."""
    mime_bundle = nodes.container(nb_element="mime_bundle")
    set_source_info(mime_bundle, source, line)
    for mime_type, content in data.items():
        mime_container = nodes.container(mime_type=mime_type)
        set_source_info(mime_container, source, line)
        mime_data = MimeData(
            mime_type,
            content,
            cell_metadata=cell_metadata,
            output_metadata=output_metadata,
            line=line,
        )
        if inline:
            _nodes = nb_renderer.render_mime_type_inline(mime_data)
        else:
            _nodes = nb_renderer.render_mime_type(mime_data)
        if _nodes:
            mime_container.extend(_nodes)
            mime_bundle.append(mime_container)
    return [mime_bundle]


class PendingGlueReference(nodes.Element):
    """A glue reference to another document."""

    @property
    def refdoc(self) -> str:
        return self.attributes["refdoc"]

    @property
    def key(self) -> str:
        return self.attributes["key"]

    @property
    def inline(self) -> bool:
        return self.attributes.get("inline", False)

    @property
    def gtype(self) -> Optional[str]:
        return self.attributes.get("gtype", None)


class PendingGlueReferenceError(Exception):
    """An error occurred while resolving a pending glue reference."""


def create_pending_glue_ref(
    document: nodes.document,
    source: str,
    line: int,
    rel_doc: str,
    key: str,
    inline: bool = False,
    gtype: Optional[str] = None,
    **kwargs: Any,
) -> PendingGlueReference:
    """Create a pending glue reference."""
    if not is_sphinx(document):
        raise PendingGlueReferenceError(
            "Pending glue references are only supported in sphinx."
        )
    env: "BuildEnvironment" = document.settings.env
    _, filepath = env.relfn2path(rel_doc, env.docname)
    refdoc = env.path2doc(filepath)
    if refdoc is None:
        raise PendingGlueReferenceError(
            f"Pending glue reference document not found: {filepath!r}."
        )
    ref = PendingGlueReference(
        refdoc=refdoc, key=key, inline=inline, gtype=gtype, **kwargs
    )
    ref.source = source
    ref.line = line
    return ref


def format_plain_text(text: str, fmt_spec: str) -> str:
    """Format plain text for display in a docutils node."""
    # literal eval, to remove surrounding quotes
    try:
        value = literal_eval(text)
    except (SyntaxError, ValueError):
        value = text
    if fmt_spec == "":
        return str(value)
    type_char = fmt_spec[-1]
    if type_char == "s":
        value = str(value)
    elif type_char in ("b", "c", "d", "o", "x", "X"):
        value = int(value)
    elif type_char in ("e", "E", "f", "F", "g", "G", "n", "%"):
        value = float(value)
    return format(value, fmt_spec)



"""Functionality for storing special data in notebook code cells,
which can then be inserted into the document body.
"""
from typing import Any, Dict, List

import IPython
from IPython.display import display as ipy_display
from nbformat import NotebookNode

from myst_nb.core.loggers import LoggerType

GLUE_PREFIX = "application/papermill.record/"


def get_glue_roles(prefix: str = "glue:") -> Dict[str, Any]:
    """Return mapping of role names to role functions."""
    from .roles import PasteMarkdownRole, PasteRoleAny, PasteTextRole

    return {
        f"{prefix}": PasteRoleAny(),
        f"{prefix}any": PasteRoleAny(),
        f"{prefix}text": PasteTextRole(),
        f"{prefix}md": PasteMarkdownRole(),
    }


def get_glue_directives(prefix: str = "glue:") -> Dict[str, Any]:
    """Return mapping of directive names to directive functions."""
    from .directives import (
        PasteAnyDirective,
        PasteFigureDirective,
        PasteMarkdownDirective,
        PasteMathDirective,
    )

    return {
        f"{prefix}": PasteAnyDirective,
        f"{prefix}any": PasteAnyDirective,
        f"{prefix}figure": PasteFigureDirective,
        f"{prefix}math": PasteMathDirective,
        f"{prefix}md": PasteMarkdownDirective,
    }


def glue(name: str, variable, display: bool = True) -> None:
    """Glue a variable into the notebook's cell metadata.

    Parameters
    ----------
    name: string
        A unique name for the variable. You can use this name to refer to the variable
        later on.
    variable: Python object
        A variable in Python for which you'd like to store its display value. This is
        not quite the same as storing the object itself - the stored information is
        what is *displayed* when you print or show the object in a Jupyter Notebook.
    display: bool
        Display the object you are gluing. This is helpful in sanity-checking the
        state of the object at glue-time.
    """
    mimebundle, metadata = IPython.core.formatters.format_display_data(variable)
    mime_prefix = "" if display else GLUE_PREFIX
    metadata["scrapbook"] = dict(name=name, mime_prefix=mime_prefix)
    ipy_display(
        {mime_prefix + k: v for k, v in mimebundle.items()}, raw=True, metadata=metadata
    )


def extract_glue_data(
    notebook: NotebookNode,
    resources: Dict[str, Any],
    source_map: List[int],
    logger: LoggerType,
) -> None:
    """Extract all the glue data from the notebook, into the resources dictionary."""
    # note this assumes v4 notebook format
    data: Dict[str, NotebookNode] = resources.setdefault("glue", {})
    for index, cell in enumerate(notebook.cells):
        if cell.cell_type != "code":
            continue
        outputs = []
        for output in cell.get("outputs", []):
            meta = output.get("metadata", {})
            if "scrapbook" not in meta:
                outputs.append(output)
                continue
            key = meta["scrapbook"]["name"]
            mime_prefix = len(meta["scrapbook"].get("mime_prefix", ""))
            if key in data:
                logger.warning(
                    f"glue key {key!r} duplicate",
                    subtype="glue",
                    line=source_map[index],
                )
            output["data"] = {k[mime_prefix:]: v for k, v in output["data"].items()}
            data[key] = output
            if not mime_prefix:
                # assume that the output is a displayable object
                outputs.append(output)
        cell.outputs = outputs



"""Module for executing notebooks."""
from __future__ import annotations

from contextlib import nullcontext, suppress
from datetime import datetime
import os
from pathlib import Path, PurePosixPath
from tempfile import TemporaryDirectory

from jupyter_cache import get_cache
from jupyter_cache.base import CacheBundleIn
from jupyter_cache.cache.db import NbProjectRecord
from jupyter_cache.executors.utils import single_nb_execution
from nbformat import NotebookNode
from typing_extensions import TypedDict

from myst_nb.core.config import NbParserConfig
from myst_nb.core.loggers import LoggerType


class ExecutionResult(TypedDict):
    """Result of executing a notebook."""

    mtime: float
    """POSIX timestamp of the execution time"""
    runtime: float | None
    """runtime in seconds"""
    method: str
    """method used to execute the notebook"""
    succeeded: bool
    """True if the notebook executed successfully"""
    error: str | None
    """error type if the notebook failed to execute"""
    traceback: str | None
    """traceback if the notebook failed"""


class ExecutionError(Exception):
    """An exception for failed execution and `execution_raise_on_error` is true."""


def execute_notebook(
    notebook: NotebookNode,
    source: str,
    nb_config: NbParserConfig,
    logger: LoggerType,
    read_fmt: None | dict = None,
) -> tuple[NotebookNode, ExecutionResult | None]:
    """Update a notebook's outputs using the given configuration.

    This function may execute the notebook if necessary, to update its outputs,
    or populate from a cache.

    :param notebook: The notebook to update.
    :param source: Path to or description of the input source being processed.
    :param nb_config: The configuration for the notebook parser.
    :param logger: The logger to use.
    :param read_fmt: The format of the input source (to parse to jupyter cache)

    :returns: The updated notebook, and the (optional) execution metadata.
    """
    # TODO should any of the logging messages be debug instead of info?

    # path should only be None when using docutils programmatically,
    # e.g. source="<string>"
    try:
        path = Path(source) if Path(source).is_file() else None
    except OSError:
        path = None  # occurs on Windows for `source="<string>"`

    exec_metadata: ExecutionResult | None = None

    # check if the notebook is excluded from execution by pattern
    if path is not None and nb_config.execution_excludepatterns:
        posix_path = PurePosixPath(path.as_posix())
        for pattern in nb_config.execution_excludepatterns:
            if posix_path.match(pattern):
                logger.info(f"Excluded from execution by pattern: {pattern!r}")
                return notebook, exec_metadata

    # 'auto' mode only executes the notebook if it is missing at least one output
    missing_outputs = (
        len(cell.outputs) == 0 for cell in notebook.cells if cell["cell_type"] == "code"
    )
    if nb_config.execution_mode == "auto" and not any(missing_outputs):
        logger.info("Skipped execution in 'auto' mode (all outputs present)")
        return notebook, exec_metadata

    if nb_config.execution_mode in ("auto", "force"):

        # setup the execution current working directory
        if nb_config.execution_in_temp:
            cwd_context = TemporaryDirectory()
        else:
            if path is None:
                raise ValueError(
                    f"source must exist as file, if execution_in_temp=False: {source}"
                )
            cwd_context = nullcontext(str(path.parent))  # type: ignore

        # execute in the context of the current working directory
        with cwd_context as cwd:
            cwd = os.path.abspath(cwd)
            logger.info(
                "Executing notebook using "
                + ("temporary" if nb_config.execution_in_temp else "local")
                + " CWD"
            )
            result = single_nb_execution(
                notebook,
                cwd=cwd,
                allow_errors=nb_config.execution_allow_errors,
                timeout=nb_config.execution_timeout,
                meta_override=True,  # TODO still support this?
            )

        if result.err is not None:
            if nb_config.execution_raise_on_error:
                raise ExecutionError(str(source)) from result.err
            msg = f"Executing notebook failed: {result.err.__class__.__name__}"
            if nb_config.execution_show_tb:
                msg += f"\n{result.exc_string}"
            logger.warning(msg, subtype="exec")
        else:
            logger.info(f"Executed notebook in {result.time:.2f} seconds")

        exec_metadata = {
            "mtime": datetime.now().timestamp(),
            "runtime": result.time,
            "method": nb_config.execution_mode,
            "succeeded": False if result.err else True,
            "error": f"{result.err.__class__.__name__}" if result.err else None,
            "traceback": result.exc_string if result.err else None,
        }

    elif nb_config.execution_mode == "cache":

        # setup the cache
        cache = get_cache(nb_config.execution_cache_path or ".jupyter_cache")
        # TODO config on what notebook/cell metadata to hash/merge

        # attempt to match the notebook to one in the cache
        cache_record = None
        with suppress(KeyError):
            cache_record = cache.match_cache_notebook(notebook)

        # use the cached notebook if it exists
        if cache_record is not None:
            logger.info(f"Using cached notebook: ID={cache_record.pk}")
            _, notebook = cache.merge_match_into_notebook(notebook)
            exec_metadata = {
                "mtime": cache_record.created.timestamp(),
                "runtime": cache_record.data.get("execution_seconds", None),
                "method": nb_config.execution_mode,
                "succeeded": True,
                "error": None,
                "traceback": None,
            }
            return notebook, exec_metadata

        if path is None:
            raise ValueError(
                f"source must exist as file, if execution_mode is 'cache': {source}"
            )

        # attempt to execute the notebook
        if read_fmt is not None:
            stage_record = cache.add_nb_to_project(str(path), read_data=read_fmt)
        else:
            stage_record = cache.add_nb_to_project(str(path))
        # TODO do in try/except, in case of db write errors
        NbProjectRecord.remove_tracebacks([stage_record.pk], cache.db)
        cwd_context = (
            TemporaryDirectory()  # type: ignore
            if nb_config.execution_in_temp
            else nullcontext(str(path.parent))
        )
        with cwd_context as cwd:
            cwd = os.path.abspath(cwd)
            logger.info(
                "Executing notebook using "
                + ("temporary" if nb_config.execution_in_temp else "local")
                + " CWD"
            )
            result = single_nb_execution(
                notebook,
                cwd=cwd,
                allow_errors=nb_config.execution_allow_errors,
                timeout=nb_config.execution_timeout,
                meta_override=True,  # TODO still support this?
            )

        # handle success / failure cases
        # TODO do in try/except to be careful (in case of database write errors?
        if result.err is not None:
            if nb_config.execution_raise_on_error:
                raise ExecutionError(str(source)) from result.err
            msg = f"Executing notebook failed: {result.err.__class__.__name__}"
            if nb_config.execution_show_tb:
                msg += f"\n{result.exc_string}"
            logger.warning(msg, subtype="exec")
            NbProjectRecord.set_traceback(stage_record.uri, result.exc_string, cache.db)
        else:
            logger.info(f"Executed notebook in {result.time:.2f} seconds")
            cache_record = cache.cache_notebook_bundle(
                CacheBundleIn(
                    notebook, stage_record.uri, data={"execution_seconds": result.time}
                ),
                check_validity=False,
                overwrite=True,
            )
            logger.info(f"Cached executed notebook: ID={cache_record.pk}")

        exec_metadata = {
            "mtime": datetime.now().timestamp(),
            "runtime": result.time,
            "method": nb_config.execution_mode,
            "succeeded": False if result.err else True,
            "error": f"{result.err.__class__.__name__}" if result.err else None,
            "traceback": result.exc_string if result.err else None,
        }

    return notebook, exec_metadata



"""notebook "pre-processing" (after execution, but before parsing)"""
from __future__ import annotations

import re
from typing import Any

from nbformat import NotebookNode

from myst_nb.core.config import NbParserConfig
from myst_nb.core.loggers import LoggerType
from myst_nb.glue import extract_glue_data


def preprocess_notebook(
    notebook: NotebookNode, logger: LoggerType, nb_config: NbParserConfig
) -> dict[str, Any]:
    """Modify notebook and resources in-place."""
    # TODO make this pluggable
    # (similar to nbconvert preprocessors, but parse config, source map and logger)

    resources: dict[str, Any] = {}

    # create source map
    source_map = notebook.metadata.get("source_map", None)
    # use 1-based indexing rather than 0, or pseudo base of the cell index
    source_map = [
        (source_map[i] if source_map else ((i + 1) * 10000)) + 1
        for i, _ in enumerate(notebook.cells)
    ]

    # coalesce_streams
    for i, cell in enumerate(notebook.cells):
        if cell.cell_type == "code":
            _callback = lambda m, t: logger.warning(  # noqa: E731
                m, subtype=t, line=source_map[i]
            )
            if nb_config.get_cell_level_config(
                "merge_streams", cell.metadata, _callback
            ):
                cell["outputs"] = coalesce_streams(cell.get("outputs", []))

    # extract all scrapbook (aka glue) outputs from notebook
    extract_glue_data(notebook, resources, source_map, logger)

    return resources


_RGX_CARRIAGERETURN = re.compile(r".*\r(?=[^\n])")
_RGX_BACKSPACE = re.compile(r"[^\n]\b")


def coalesce_streams(outputs: list[NotebookNode]) -> list[NotebookNode]:
    """Merge all stream outputs with shared names into single streams.

    This ensure deterministic outputs.

    Adapted from:
    https://github.com/computationalmodelling/nbval/blob/master/nbval/plugin.py.
    """
    if not outputs:
        return []

    new_outputs = []
    streams: dict[str, NotebookNode] = {}
    for output in outputs:
        if output["output_type"] == "stream":
            if output["name"] in streams:
                streams[output["name"]]["text"] += output["text"]
            else:
                new_outputs.append(output)
                streams[output["name"]] = output
        else:
            new_outputs.append(output)

    # process \r and \b characters
    for output in streams.values():
        old = output["text"]
        while len(output["text"]) < len(old):
            old = output["text"]
            # Cancel out anything-but-newline followed by backspace
            output["text"] = _RGX_BACKSPACE.sub("", output["text"])
        # Replace all carriage returns not followed by newline
        output["text"] = _RGX_CARRIAGERETURN.sub("", output["text"])

    # We also want to ensure stdout and stderr are always in the same consecutive order,
    # because they are asynchronous, so order isn't guaranteed.
    for i, output in enumerate(new_outputs):
        if output["output_type"] == "stream" and output["name"] == "stderr":
            if (
                len(new_outputs) >= i + 2
                and new_outputs[i + 1]["output_type"] == "stream"
                and new_outputs[i + 1]["name"] == "stdout"
            ):
                stdout = new_outputs.pop(i + 1)
                new_outputs.insert(i, stdout)

    return new_outputs
